/****************************************************************************
*
*  Local Includes
*
***************************************************************************/

#define REG(x) register

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <stdarg.h>
#include <ctype.h>
#include <errno.h>
#include <sys/types.h>
#include <unistd.h>
#include <string.h>

#include "stdcode.h"
#include "FPL.h"
#include "FPL_protos.h"
#include "reference.h"
#include "os.h"
#include "mystring.h"
#include "sprintf.h"
#include "fplerr.h"
#include "files.h"
#include "urlget.h"

#if 0
#include "user.h"
#include "comm.h"
#include "colors.h"
#include "message.h"
#include "system.h"
#include "serial.h"
#include "flink.h"
#endif

#define VERSION "0.6"

#define CALLER
#define ASM

#define FPLSendInt( xxx )     fplSendTags( anchor, FPLSEND_INT, xxx, \
					   FPLSEND_DONE )
#define FPLSendString( xxx )  fplSendTags( anchor, FPLSEND_STRING, xxx, \
					   FPLSEND_STRLEN, strlen( xxx ), \
					   FPLSEND_DONE)

#define FPLSendStringLen( xxx, lll )  fplSendTags( anchor, FPLSEND_STRING, xxx, \
					   FPLSEND_STRLEN, lll, \
					   FPLSEND_DONE)

/****************************************************************************
*
*  Local Function Declarations
*
***************************************************************************/

static long     fpl_Func (struct fplArgument *);
static int 	sortCmp ( const void* e1, const void* e2 );
static int 	sortCmpi( const void* e1, const void* e2 );

/****************************************************************************
*
*  Global Variables
*
***************************************************************************/

void* key;
int   FplErr_I;
int   RunSafe_I = 0;
char  SafePath_AC[256];

static int Argc_I;
static char** Argv_PPC;
static char  DayName_AAC[3][10];
static char  MonthName_AAC[12][10];

typedef struct {
    char* name;
    int   id;
    int   rtype;
    char* params;
} fplFunc;

enum myfunctions {

  /* programmer fuctions*/
  FN_GETINT, FN_GETSTRING, FN_GETCHAR,
  FN_TOUPPER, FN_TOLOWER,
  FN_RUNFILE, FN_RUNSAFE,
  FN_PRINTF, 
  FN_SYSTEM, FN_POPEN,
  FN_LOADSTRING, FN_SAVESTRING, FN_LOADARRAY, FN_SAVEARRAY,
  FN_STR2ARRAY, FN_ARRAY2STR,

  FN_TRANSLATE, FN_FAPPEND, FN_FPREPEND,FN_FEXISTS, FN_GETENV,
  FN_TOKSTR, FN_SETDATESTR, FN_PARSECMND, FN_RANDOM, FN_SLEEP, FN_SORT,
  FN_URL2TEXT, FN_TEXT2URL, FN_GETURLVAR, FN_URLGET,

/*
  FN_STR2ARRAY, FN_ARRAY2STR,
  */

  /* user functions */
  FN_GETUID, FN_GETGID, FN_GETPID,

  /* file functions */
  FN_FILEINFO, FN_FILEFLAG, FN_SPLITPATH, FN_REMOVE, FN_RENAME,
  FN_FFINDFIRST, FN_FFINDNEXT,
  FN_CHDIR, FN_CHOWN, FN_CHMOD,

  /* time functions */
  FN_TIME, FN_TIME2STR, FN_STR2TIME, FN_TIMEPART, FN_STRFTIME,

  /* argv func */
  FN_ARGV,

  VR_ARGC, VR_ERRNO,
  VR_VERSION,VR_FPLVERSION,VR_SAFEMODE,

  FN_NOP

};

static fplFunc vars[] = {
    /* variables */
    { "argc",        VR_ARGC,      'I',  0 },
    { "errno",       VR_ERRNO,     'I',  0 },
    { "fsh_version", VR_VERSION,   'S',  0 },
    { "fpl_version", VR_FPLVERSION,'S',  0 },
    { "fsh_safemode",VR_SAFEMODE,  'I',  0 },

    { NULL, 0, 0, NULL }
};

static fplFunc funcs[] = {
    { "getline",   FN_GETSTRING, 'S', "si" },
    { "getint",    FN_GETINT,    'I', "si" },
    { "getchar",   FN_GETCHAR,   'I', "si" },
    { "loadstring",FN_LOADSTRING,'S', "S" },
    { "savestring",FN_SAVESTRING,'I', "SSi" },
    { "toupper",   FN_TOUPPER,   'I', "I" },
    { "tolower",   FN_TOLOWER,   'I', "I" },
    { "runfile",   FN_RUNFILE,   'I', "S" },
    { "runsafe",   FN_RUNSAFE,   'I', "SS" },
    { "printf",    FN_PRINTF,    'I', "So>" },
    { "system",    FN_SYSTEM,    'I', "S" },
    { "popen",     FN_POPEN,     'S', "Ss" },
    { "getenv",    FN_GETENV,    'S', "S" },
    { "tokstr",    FN_TOKSTR,    'S', "SIs" },
    { "setdatestr",FN_SETDATESTR,'I', "BB" },
    { "parsecmnd", FN_PARSECMND, 'S', "SB" },
    { "random",    FN_RANDOM,    'I', "" },
    { "sleep",     FN_SLEEP,     'I', "I" },
    { "sort",	   FN_SORT,      'I', "Bi" },
    { "loadarray", FN_LOADARRAY, 'I', "SB" },
    { "savearray", FN_SAVEARRAY, 'I', "BSi" },
    { "str2array", FN_STR2ARRAY, 'I', "SBi" },
    { "array2str", FN_ARRAY2STR, 'S', "Bi" },
    { "url2text",  FN_URL2TEXT,  'S', "S" },
    { "text2url",  FN_TEXT2URL,  'S', "S" },
    { "geturlvar", FN_GETURLVAR, 'S', "SS" },
    { "urlget",    FN_URLGET,    'S', "Ssi" },

    /* user functions */
    { "getpid",    FN_GETPID,    'I', "" },
    { "getuid",    FN_GETUID,    'I', "" },
    { "getgid",    FN_GETGID,    'I', "" },
  

    /* file functions */
    { "fappend",   FN_FAPPEND,   'I', "SS" },
    { "fprepend",  FN_FPREPEND,  'I', "SS" },
    { "fexists",   FN_FEXISTS,   'I', "S" },
    { "fileinfo",  FN_FILEINFO,  'I', "SS" },
    { "fileflag",  FN_FILEFLAG,  'S', "S" },
    { "splitpath", FN_SPLITPATH, 'S', "SS" },
    { "remove",    FN_REMOVE,    'I', "S" },
    { "rename",    FN_RENAME,    'I', "SS" },
    { "chdir",     FN_CHDIR,     'I', "S" },
#if 0
    { "chmod",     FN_CHMOD,     'I', "SS" },
    { "chown",     FN_CHOWN,     'I', "SSs" },
    { "ffindfirst",FN_FFINDFIRST,'S', "Ss" },
    { "ffindnext", FN_FFINDNEXT, 'S', "" },
#endif

    /* time functions */
    { "time",   FN_TIME,   'I', "" },
    { "timepart",  FN_TIMEPART,  'I', "IS" },
    { "time2str",  FN_TIME2STR,  'S', "I" },
    { "str2time",  FN_STR2TIME,  'I', "S" },
    { "strftime",  FN_STRFTIME,  'S', "Si" },
    { "argv",      FN_ARGV,      'S', "I" },

    { NULL, 0, 0, NULL }
};


t_PartId FPL_FileInfo_AAC[] = {
    "size",
    "year",
    "month",
    "day",
    "hour",
    "minute",
    "second",
    "time"
};

t_PartId FPL_SplitPath_AAC[] = {
    "path",
    "file"
};

t_PartId FPL_TimePart_AAC[] = {
    "year",
    "month",
    "day",
    "hour",
    "minute",
    "second",
    "wday"
};

int FPL_Init( char* StartFile_PC );

int main( int argc, char* argv[] )
{
    int rc;

    if ( 1 == argc ) {
        printf("FrexxShell " VERSION "\n");
        printf("usage: %s [fpl file]\n", argv[0] );
        return -1;
    }

    Argc_I = argc - 1;
    Argv_PPC = argv;

    rc = FPL_Init( argv[1] );

    return 0;
}
/****************************************************************************
*
*  Function:    FPL_Init
*  Description: Sets up the FPL environment and starts execution of the FPL
*               files.
*
***************************************************************************/

int FPL_Init( char* StartFile_PC )
{
  long end=0;
  int ms,i;

  OS_Milliseconds( &ms );
  srand( ms );

  key=fplInitTags(fpl_Func, 
		  FPLTAG_CACHEALLFILES, 1, 
		  FPLTAG_REREAD_CHANGES, 1, 
		  FPLTAG_DONE );
  if (!key) {
      printf("Error initializing FPL!\n");
      return -1;
  }

#if 0
  long Version_I,Revision_I,rc;
  rc = fplSendTags( key, FPLSEND_GETVERSION,  &Version_I,
                         FPLSEND_GETREVISION, &Revision_I, FPLSEND_DONE );
  if ( rc )
      RET_ERR( "Error getting FPL version", (int)rc );

  printf("Using FPL.DLL version %ld.%ld.\n",Version_I,Revision_I);
#endif

  /* setup functions */
  for (i=0;funcs[i].name;i++)
      fplAddFunction( key, funcs[i].name, funcs[i].id, 
                      funcs[i].rtype, funcs[i].params, NULL );

  /* setup variables */
  for (i=0;vars[i].name;i++)
      fplAddVariable( key, vars[i].name, vars[i].id, 
                      vars[i].rtype, vars[i].params, NULL );

  /* run FPL file */
  end = fplExecuteFile(key,StartFile_PC,NULL);

  switch ( end ) {
    case FPL_OK:
    case FPL_EXIT_OK:
      break;

    case FPL_OPEN_ERROR:
      printf("FPL file not found: %s\n",StartFile_PC );
      break;

    default:
      printf("FPL reported internal error %ld\n",end );
      break;
  }

  fplFree(key); /* free all shit FPL uses internally */

  return end;
}


/****************************************************************************
*
*  Function:    FPL_FixPath
*  Description: Fix path for "safe" mode
*
***************************************************************************/
char* FPL_FixPath( char* File_PC )
{
    static char String_AC[256];

    sprintf(String_AC,"%s/%s",SafePath_AC,File_PC);
    return String_AC;
}

int sortCmp( const void* e1, const void* e2 )
{
    char** s1 = (char**) e1;
    char** s2 = (char**) e2;

    return strcmp(*s1,*s2);
}

int sortCmpi( const void* e1, const void* e2 )
{
    char** s1 = (char**) e1;
    char** s2 = (char**) e2;

    return strcasecmp(*s1,*s2);
}

/****************************************************************************
*
*  Function:    fpl_Func
*  Description: Interface function for FPL. Contains *all* FPL functions.
*
***************************************************************************/

#undef CALL
#define CALL( function )  do {if ( OK != function ) return OK;} while(0)

#define CHK_SAFE  if (RunSafe_I) break

long ASM fpl_Func(REG(a0) struct fplArgument *arg)
{
    int ret = OK;
    long col;
    char *name;
    char *string;
    void *anchor=arg->key;
    char* String_PC;
    char* TmpStr_PC;
    char  TmpStr_AC[ 1024 ];
    int	  TmpInt_I=0;

    t_Bool SendInt_B=FALSE, SendString_B=FALSE;

    switch( arg->ID ) {

	/*****************  FPL internal functions ********************/

      case FPL_GENERAL_ERROR: {
	  char  buffer[FPL_ERRORMSG_LENGTH];
	  char* TmpStr_PC;
	  int   TmpInt_I;
	  FILE* File_PS;
		  
	  TmpInt_I = (long) arg->argv[0];
	  TmpStr_PC = fplGetErrorMsg( arg->key, (long)TmpInt_I, buffer );
	  
	  fplSendTags(anchor,
		      FPLSEND_GETVIRLINE, &col,
		      FPLSEND_GETVIRFILE, &name,
		      FPLSEND_DONE);
	  
          printf("\n>>> FPL error in %s, line %ld - %s\n\n",
                 name,col,TmpStr_PC);

	  /* log the error */
	  File_PS = fopen( "fplerr.log", "a" );
	  if (File_PS) {
	      time_t Time_I = time(NULL);
	      struct tm* Time_PS = localtime( &Time_I );
	      fprintf( File_PS,
		      "%04d-%02d-%02d,%02d:%02d:%02d%c%s, line %ld - %s\n",
		      Time_PS->tm_year + 1900,
		      Time_PS->tm_mon + 1,
		      Time_PS->tm_mday,
		      Time_PS->tm_hour,
		      Time_PS->tm_min,
		      Time_PS->tm_sec,
		      RunSafe_I?'*':' ',
		      name,col,TmpStr_PC);
	      fclose(File_PS);
	  }

	  if (RunSafe_I)
	      ret = OK;
	  else
	      ret = ERROR;
      }

      case FPL_UNKNOWN_FUNCTION:
	col=22;			/* only to breakpoint */
	break;
	
      case FPL_COMPILE: /* for experiments! */
	if(arg->format[0]==FPL_STRARG)
	    printf("%s", (char*)arg->funcdata);
	else
	    printf("%d", (int)arg->funcdata);
	break;
	
	/*****************  FrexxLink functions ********************/

      case FN_PRINTF:
	if(!FPL_STRLEN(arg->argv[0]))
	    break;
	String_PC = NULL;
        STR_Sprintf( &String_PC, arg->argv[0], arg->argv, arg->format );
        printf( "%s", String_PC );
        fflush(stdout);
	free( String_PC );
	break;

      case FN_GETENV:
	String_PC = getenv( arg->argv[0] );
	if ( String_PC )
	    ret = FPLSendString( String_PC );
	break;

      case FN_FAPPEND: {
	FILE* File_PS;

	CHK_SAFE;
	File_PS = fopen( arg->argv[0], "a" );
	if ( !File_PS ) {
	    fprintf(stderr, "Can't open '%s'. %s\n", 
		    (char*)arg->argv[0], strerror(errno) );
	    FplErr_I = FPLERR_NO_FILE;
	    break;
	}
	fprintf( File_PS, "%s", (char*)arg->argv[1] );
	fclose( File_PS );
	break;
      }

      case FN_FPREPEND:
	CHK_SAFE;
	FIL_Prepend(arg->argv[0],arg->argv[1]);
	break;

      case FN_FEXISTS: {
	FILE* File_PS;

	File_PS = fopen( arg->argv[0], "r" );
	if ( File_PS ) {
	    TmpInt_I = 1;
	    fclose( File_PS );
	}
	else
	    TmpInt_I = 0;
	SendInt_B = TRUE;
	break;
      }

      case FN_LOADSTRING:
	if ( RunSafe_I )
	    TmpStr_PC = FPL_FixPath(arg->argv[0]);
	else
	    TmpStr_PC = arg->argv[0];
	STR_LoadString( TmpStr_PC, &String_PC );
	if ( String_PC ) {
	    ret = FPLSendString( String_PC );
	    free( String_PC );
	}
	else
	    ret = FPLSendString( "" );
	break;

      case FN_SAVESTRING:
	if ( RunSafe_I )
	    TmpStr_PC = FPL_FixPath(arg->argv[1]);
	else
	    TmpStr_PC = arg->argv[1];

        TmpInt_I = FPL_STRLEN( arg->argv[0] );

	STR_SaveString( arg->argv[0], TmpStr_PC, TmpInt_I );
	break;

      case FN_LOADARRAY: {
	  struct fplRef refStruct;
	  long l[2];
	  unsigned long tags[5];
	  int  Items_I,i;
	  long ll;
	  char** List_PPC;

	  TmpInt_I = -1;
	  SendInt_B = TRUE;

	  if ( RunSafe_I )
	      TmpStr_PC = FPL_FixPath(arg->argv[0]);
	  else
	      TmpStr_PC = arg->argv[0];

	  CALL( STR_LoadArray( TmpStr_PC, &List_PPC, &Items_I ));
	  if (!List_PPC) {
	      TmpInt_I = 0;
	      break;
	  }

	  ll = Items_I;
	  refStruct.ArraySize = &ll;
	  refStruct.Dimensions = 1;

	  tags[0] = (long) FPLREF_ARRAY_RESIZE;
	  tags[1] = (long) &refStruct;
	  tags[2] = (long) FPLREF_DONE;
	  fplReference(key, arg->argv[1], tags );

	  tags[0] = (long) FPLREF_ARRAY_ITEM;
	  tags[1] = (long) &l;
	  tags[2] = (long) FPLREF_SET_MY_STRING;
	  tags[3] = (long) &string;
	  tags[4] = (long) FPLREF_DONE;

	  l[1] = -1;
	  for (i=0; i<Items_I; i++) {
	      l[0] = i;
 	      tags[3] = (long) List_PPC[i];
	      fplReference( key, arg->argv[1], tags );
	  }
	  free(List_PPC);
	  TmpInt_I = Items_I;
      }
	break;

      case FN_SAVEARRAY: {
	  struct fplRef refStruct;
	  long l[2];
	  unsigned long tags[5];
	  int  items,i;
	  FILE* File_PS;

	  tags[0] = (long) FPLREF_ARRAY_INFO;
	  tags[1] = (long) &refStruct;
	  tags[2] = (long) FPLREF_DONE;
	  
	  fplReference( key, arg->argv[0], tags );
	  items = *refStruct.ArraySize;
	  if (!items) {
	      RET_ERR("No elements in array!",0);
	      break;
	  }

	  /* angav anropet antal rader? */
	  if ( arg->argc > 2 ) {
	      TmpInt_I = (int) arg->argv[2];
	      if ( TmpInt_I < items )
		  items = TmpInt_I;
	  }

	  if ( RunSafe_I )
	      TmpStr_PC = FPL_FixPath(arg->argv[1]);
	  else
	      TmpStr_PC = arg->argv[1];

	  File_PS = fopen( TmpStr_PC, "w" );
	  if ( !File_PS ) {
	      sprintf(TmpStr_AC, "Can't open '%s'. %s\n", 
		      (char*)arg->argv[0], strerror(errno) );
	      RET_ERR(TmpStr_AC,errno);
	  }

	  tags[0] = (long) FPLREF_ARRAY_ITEM;
	  tags[1] = (long) &l;
	  tags[2] = (long) FPLREF_GET_STRING;
	  tags[3] = (long) &string;
	  tags[4] = (long) FPLREF_DONE;

	  l[1] = -1;
	  for (i=0; i<items; i++) {
	      int len;
	      l[0] = i;
	      fplReference(key, arg->argv[0], tags );
	      len = FPL_STRLEN( string );
	      fwrite( string, len, 1, File_PS );
	  }

	  fclose( File_PS );
      }
	break;

      case FN_GETSTRING: {
          char* Line_PC;
          int   Length_I = 0;
          t_Bool More_B;
          t_RetCode rc;
          
          if ( arg->argc )
              printf("%s", (char*) arg->argv[0] );
          if ( arg->argc > 1 )
              TmpInt_I = (int) (arg->argv[1]);
          else
              TmpInt_I = -1;

          rc = STR_ReadLine( stdin, &Line_PC, &Length_I, &More_B );
          if ( OK == rc ) {
              ret = FPLSendString( Line_PC );
              free( Line_PC );
          }
          else
              ret = FPLSendString( "" );
          break;
      }

      case FN_GETINT:
	if( arg->argc )
	    printf("%s", (char*) arg->argv[0] );
	if ( arg->argc > 1 )
	    TmpInt_I = (int) (arg->argv[1]);
	else
	    TmpInt_I = -1;
	fgets( TmpStr_AC, 1024, stdin );

	TmpInt_I  = atoi( TmpStr_AC );
	SendInt_B = TRUE;
	break;

      case FN_GETCHAR: {
	  if ( arg->argc )
	      printf( "%s", (char*) arg->argv[0] );
	  if ( arg->argc > 1 )
	      TmpInt_I = (int) (arg->argv[1]);
	  else
	      TmpInt_I = -1;
	  TmpInt_I = getc( stdin );

	  SendInt_B = TRUE;
	  break;
      }
	
      case FN_TOKSTR:
	if ( arg->argc > 2 )
	    strcpy( TmpStr_AC, arg->argv[2]);
	else
	    strcpy( TmpStr_AC, ";" );

	STR_TokStr( arg->argv[0], (int) arg->argv[1], TmpStr_AC, &String_PC );
	if (String_PC)
	    ret=FPLSendString( String_PC );
	else
	    ret=FPLSendString("");
	free( String_PC );
	break;

      case FN_SORT: {
	struct fplRef refStruct;
	long l[2];
	unsigned long tags[5];
	int  items,i;
	char** list;

	tags[0] = (long) FPLREF_ARRAY_INFO;
	tags[1] = (long) &refStruct;
	tags[2] = (long) FPLREF_DONE;

	fplReference( key, arg->argv[0], tags );
	items = *refStruct.ArraySize;
	if (!items)
	    break;
	list = (char**) malloc( items * sizeof( char* ) );

	tags[0] = (long) FPLREF_ARRAY_ITEM;
	tags[1] = (long) &l;
	tags[2] = (long) FPLREF_GET_STRING;
	tags[3] = (long) &string;
	tags[4] = (long) FPLREF_DONE;

	l[1] = -1;
	for (i=0; i<items; i++) {
	    int len;
	    l[0] = i;
	    fplReference(key, arg->argv[0], tags );
	    len = FPL_STRLEN( string );
	    list[i] = (char*) fplAllocString( key, len );
	    memcpy(list[i], string, len);
	}

	if ( ( arg->argc > 1 ) && ( !(int)arg->argv[1] ) )
	    qsort(list,items,sizeof(char*),sortCmpi);
	else
	    qsort(list,items,sizeof(char*),sortCmp);

	tags[2] = FPLREF_SET_STRING;
	for (i=0; i<items; i++) {
	    l[0] = i;
	    tags[3] = (long) list[i];

	    fplReference(key, arg->argv[0], tags );
	}
	free( list );
      }
	break;
	

      case FN_SETDATESTR: {
	long l[2];
	long tags[5];
	char* string;

	tags[0] = (long) FPLREF_ARRAY_ITEM;
	tags[1] = (long) &l;
	tags[2] = (long) FPLREF_GET_STRING;
	tags[3] = (long) &string;
	tags[4] = (long) FPLREF_DONE;

	l[1] = -1;
	for (TmpInt_I=0;TmpInt_I<3;TmpInt_I++) {
	    l[0] = TmpInt_I;

	    fplReference( key, arg->argv[0], (unsigned long*) tags);
	    if (strlen(string))
		strcpy(DayName_AAC[TmpInt_I],string);
	    else
		DayName_AAC[TmpInt_I][0] = 0;	
	}

	for (TmpInt_I=0;TmpInt_I<12;TmpInt_I++) {
	    l[0] = TmpInt_I;
	    fplReference( key, arg->argv[1], (unsigned long*) tags);

	    if (strlen(string))
		strcpy(MonthName_AAC[TmpInt_I],string);
	    else
		MonthName_AAC[TmpInt_I][0] = 0;
	}

      }
	break;

      case FN_PARSECMND: {
	struct fplRef refStruct;
	long l[2];
	unsigned long tags[5];
	char* string;
	char  tmpStr[10];
	int  commands,hits=0;

	tags[0] = (long) FPLREF_ARRAY_INFO;
	tags[1] = (long) &refStruct;
	tags[2] = (long) FPLREF_DONE;

	fplReference( key, arg->argv[1], tags );
	commands = *refStruct.ArraySize;

	tags[0] = (long) FPLREF_ARRAY_ITEM;
	tags[1] = (long) &l;
	tags[2] = (long) FPLREF_GET_STRING;
	tags[3] = (long) &string;
	tags[4] = (long) FPLREF_DONE;

	TmpStr_AC[0] = 0;

	l[1] = -1;
	for (TmpInt_I=0; TmpInt_I<commands; TmpInt_I++) {
	    l[0] = TmpInt_I;

	    fplReference( key, arg->argv[1], tags);

	    if ( STR_Match( string, arg->argv[0] ) ) {
		sprintf(tmpStr,"%d,",TmpInt_I);
		strcat(TmpStr_AC,tmpStr);
		hits++;
	    }
	}
	if (hits)
	    SendString_B = TRUE;
	else
	    SendString_B = FALSE;
      }
	break;

      case FN_RANDOM:
	TmpInt_I = rand() % 1000;
	SendInt_B = TRUE;
	break;

      case FN_URL2TEXT:
          TmpStr_PC = STR_url2text( arg->argv[0] );
          if (TmpStr_PC) {
              FPLSendString(TmpStr_PC);
              free(TmpStr_PC);
          }
          else
              FPLSendString("");
          break;

      case FN_TEXT2URL:
          TmpStr_PC = STR_text2url( arg->argv[0] );
          if (TmpStr_PC) {
              FPLSendString(TmpStr_PC);
              free(TmpStr_PC);
          }
          else
              FPLSendString("");
          break;

      case FN_GETURLVAR:
          TmpStr_PC = STR_GetURL( arg->argv[0], arg->argv[1] );
          if (TmpStr_PC) {
              FPLSendString(TmpStr_PC);
              free(TmpStr_PC);
          }
          else
              FPLSendString("");
          break;

      case FN_URLGET: {
          long size;
          long conf = CONF_MUTE;
          unsigned short port = 1080;
          char* proxy = NULL;
          char* url = (char*) arg->argv[0];

          if ( arg->argc > 1 ) {
              conf |= CONF_PROXY;
              proxy = (char*) arg->argv[1];
              if ( arg->argc > 2 ) {
                  conf |= CONF_PORT;
                  port = (int) arg->argv[2];
              }
          }

          TmpInt_I = urlget( &TmpStr_PC, &size, url, port, proxy, conf,NULL,NULL );
          if ( !TmpInt_I )
              FPLSendStringLen( TmpStr_PC, size );
          else
              FPLSendString( "" );

          UrlCleanup();
                  
          break;
      }

      case FN_STR2ARRAY: {
	  struct fplRef refStruct;
	  long l[2];
	  unsigned long tags[5];
	  int  Items_I,i,Separator_I;
	  long ll;
	  char** List_PPC;

	  TmpInt_I = -1;
	  SendInt_B = TRUE;

          if ( arg->argc > 2 )
              Separator_I = (int) arg->argv[2];
          else
              Separator_I = '\n';

	  CALL( STR_SplitString( arg->argv[0], Separator_I,
                                 &List_PPC, &Items_I ) );
	  if (!List_PPC) {
	      TmpInt_I = 0;
	      break;
	  }

	  ll = Items_I;
	  refStruct.ArraySize = &ll;
	  refStruct.Dimensions = 1;

	  tags[0] = (long) FPLREF_ARRAY_RESIZE;
	  tags[1] = (long) &refStruct;
	  tags[2] = (long) FPLREF_DONE;
	  fplReference(key, arg->argv[1], tags );

	  tags[0] = (long) FPLREF_ARRAY_ITEM;
	  tags[1] = (long) &l;
	  tags[2] = (long) FPLREF_SET_MY_STRING;
	  tags[3] = (long) &string;
	  tags[4] = (long) FPLREF_DONE;

	  l[1] = -1;
	  for (i=0; i<Items_I; i++) {
	      l[0] = i;
 	      tags[3] = (long) List_PPC[i];
	      fplReference( key, arg->argv[1], tags );
	  }
	  free(List_PPC);
	  TmpInt_I = Items_I;
      }
	break;


      case FN_ARRAY2STR: {
	  struct fplRef refStruct;
	  long l[2];
	  unsigned long tags[5];
	  int  items,i,sep;

          int moffset = 0;
          int msize = 256;
          char* mptr = NULL;

	  tags[0] = (long) FPLREF_ARRAY_INFO;
	  tags[1] = (long) &refStruct;
	  tags[2] = (long) FPLREF_DONE;
	  
	  fplReference( key, arg->argv[0], tags );
	  items = *refStruct.ArraySize;
	  if (!items) {
	      RET_ERR("No elements in array!",0);
	      break;
	  }

          if ( arg->argc > 1 )
              sep = (int)arg->argv[1];
          else
              sep = '\n';

          mptr = (char*) malloc( msize );
          if (!mptr)
              break;

	  tags[0] = (long) FPLREF_ARRAY_ITEM;
	  tags[1] = (long) &l;
	  tags[2] = (long) FPLREF_GET_STRING;
	  tags[3] = (long) &string;
	  tags[4] = (long) FPLREF_DONE;

	  l[1] = -1;
	  for (i=0; i<items; i++) {
	      int len;
	      l[0] = i;
	      fplReference(key, arg->argv[0], tags );
	      len = FPL_STRLEN( string );

              if ( (len + moffset) > msize ) {
                  msize *= 4;
                  mptr = realloc( mptr, msize );
                  if (!mptr)
                      return -1;
              }
              memcpy( mptr+moffset, string, len );
              mptr[moffset + len] = sep;
              moffset += len + 1;
	  }
          mptr[moffset] = 0;
          FPLSendString( mptr );
          free( mptr );
      }
	break;



	/*
	 *  File functions
	 */

      case FN_FFINDFIRST:
	CHK_SAFE;
	TmpStr_AC[0] = 0;
	if ( arg->argc == 2 )
	    String_PC = arg->argv[1];
	else
	    String_PC = NULL;
	CALL( OS_FindFirst( arg->argv[0], String_PC, TmpStr_AC ) );
	if (TmpStr_AC[0])
	    SendString_B = TRUE;
	break;

      case FN_FFINDNEXT:
	CHK_SAFE;
	TmpStr_AC[0] = 0;
	CALL( OS_FindFirst( NULL,NULL,TmpStr_AC ) );
	if (TmpStr_AC[0])
	    SendString_B = TRUE;
	break;

      case FN_REMOVE:
	CHK_SAFE;
	if(remove( arg->argv[0] ))
	    TmpInt_I = errno;
	else
	    TmpInt_I = 0;
	break;

      case FN_RENAME:
	CHK_SAFE;
	if(rename( arg->argv[0],arg->argv[1] ))
	    TmpInt_I = errno;
	else
	    TmpInt_I = 0;
	break;

      case FN_CHDIR:
	CHK_SAFE;
	if (chdir( arg->argv[0] ))
	    TmpInt_I = errno;
	else
	    TmpInt_I = 0;
	break;

      case FN_FILEINFO: {
	  t_FileInfo	Info_S;
	  
	  CHK_SAFE;
	  SendInt_B = TRUE;

	  TmpInt_I = STR_WhatString( arg->argv[1], FPL_FileInfo_AAC );

	  if ( -1 == TmpInt_I )
	      break;

	  OS_FileInfo( arg->argv[0], &Info_S );
	  if ( Info_S.Missing_B ) {
	      TmpInt_I = -1;
	      break;
	  }
	  
	  switch ( TmpInt_I ) {
	    case 0: TmpInt_I = Info_S.Size_I;		break;
	    case 1: TmpInt_I = Info_S.Time_S.Year_I;	break;
	    case 2: TmpInt_I = Info_S.Time_S.Month_I;	break;
	    case 3: TmpInt_I = Info_S.Time_S.DayOfMonth_I; break;
	    case 4: TmpInt_I = Info_S.Time_S.Hour_I; 	break;
	    case 5: TmpInt_I = Info_S.Time_S.Minute_I; 	break;
	    case 6: TmpInt_I = Info_S.Time_S.Second_I; 	break;
	    case 7: TmpInt_I = Info_S.Time_S.Time_I;	break;

	    default:
	      SendInt_B = FALSE;
	      RET_ERR( "Unknown fileinfo() part no", (int) (arg->argv[1]) );
	  }
      }
	break;

      case FN_SPLITPATH: {
	TmpInt_I = STR_WhatString( arg->argv[1], FPL_SplitPath_AAC );

	strcpy( TmpStr_AC, arg->argv[0] );

	String_PC = strrchr( TmpStr_AC, '/' );
	if (String_PC)
	    *String_PC = 0;
	else
	    break;

	switch ( TmpInt_I ) {
	  case 0:
	    SendString_B = TRUE;
	    break;

	  case 1:
	    String_PC++;
	    ret = FPLSendString( String_PC );
	    break;

	  case -1:
	    sprintf( TmpStr_AC, "Unknown splitpath() part id: %s",
		     (char*) arg->argv[0] );
	    RET_ERR( TmpStr_AC, 0 );
	}
	break;
      }

      case FN_GETPID:
	TmpInt_I = getpid();
	SendInt_B = TRUE;
	break;

      case FN_GETUID:
	TmpInt_I = getuid();
	SendInt_B = TRUE;
	break;

      case FN_GETGID:
        TmpInt_I = getgid();
	SendInt_B = TRUE;
	break;

      case FN_RUNFILE:
	CHK_SAFE;
	ret = fplExecuteFile(anchor, arg->argv[0], NULL);
	break;
	
      case FN_RUNSAFE:
	CHK_SAFE;
	RunSafe_I++;

	strncpy(SafePath_AC,arg->argv[1],256);
	fplExecuteFileTags(anchor, arg->argv[0], 
			   FPLTAG_CACHEFILE, FPLCACHE_NONE,
			   FPLTAG_ISOLATE,   1, 
			   NULL);
	RunSafe_I--;
	break;
	
      case FN_TOUPPER:
	TmpInt_I = STR_ToUpper( (int)(arg->argv[0]) );
	SendInt_B = TRUE;
	break;

      case FN_TOLOWER:
	TmpInt_I = STR_ToLower( (int)(arg->argv[0]) );
	SendInt_B = TRUE;
	break;
	
      case FN_SLEEP:
	TmpInt_I = (int) arg->argv[0];
	if ( RunSafe_I && ( TmpInt_I > 10 ) )
	    TmpInt_I = 10;
	sleep( TmpInt_I );
	break;


	/*
	 *  Misc functions
	 */

      case FN_SYSTEM:
	CHK_SAFE;
	TmpInt_I = system( arg->argv[0] );
	SendInt_B = TRUE;
	break;

      case FN_POPEN: {
          FILE* Pipe_PS;
          char* Data_PC;
          int   Size_I=256;
          ulong Offset_I=0;

          CHK_SAFE;

          /* stdout arg? */
	  Pipe_PS = popen( arg->argv[0], "r" );

          if (Pipe_PS) {
	      Data_PC = (char*) malloc( Size_I );
              while(!feof(Pipe_PS)) {
                  int x;
                  x=fread( Data_PC + Offset_I, 1, Size_I - Offset_I, Pipe_PS );
                  Offset_I += x;
                  if ( Offset_I == Size_I ) {
                      Size_I *= 4;
                      Data_PC = realloc( Data_PC, Size_I );
                  }
              }
              ret = FPLSendStringLen( Data_PC, Offset_I );
              free( Data_PC );
          }
          break;
      }

      case FN_TIME:
	TmpInt_I = time( NULL );
	SendInt_B = TRUE;
	break;

      case FN_TIMEPART: {
	  struct tm* Time_PS;
	  time_t     Time_I = (unsigned int) (arg->argv[0]);
	  
	  TmpInt_I = STR_WhatString( arg->argv[1], FPL_TimePart_AAC );
	  if ( -1==TmpInt_I ) {
	      fprintf( stderr, "\nUnknown time part: '%s'.\n", (char*) arg->argv[1] );
	      break;
	  }

	  if ( Time_I>0 ) {
	      Time_PS = localtime( &Time_I );
	      switch ( TmpInt_I ) {
		case 0: TmpInt_I = Time_PS->tm_year+1900; break;
		case 1: TmpInt_I = Time_PS->tm_mon + 1; break;
		case 2: TmpInt_I = Time_PS->tm_mday; break;
		case 3: TmpInt_I = Time_PS->tm_hour; break;
		case 4: TmpInt_I = Time_PS->tm_min ; break;
		case 5: TmpInt_I = Time_PS->tm_sec ; break;
		case 6: 
		  TmpInt_I = Time_PS->tm_wday - 1; 
		  if(TmpInt_I<0)
		      TmpInt_I = 6;
		  break;
	      }
	  }
	  else
	      TmpInt_I = 0;
	  SendInt_B = TRUE;
	  break;
      }

      case FN_TIME2STR: {
	  struct tm* Time_PS;
	  time_t     Time_I;

	  Time_I = (int) (arg->argv[0]);

	  if (Time_I<1)
	      strcpy( TmpStr_AC, "<null time>" );
	  else {
	      Time_PS = localtime( &Time_I );
	      sprintf( TmpStr_AC, "%04d-%02d-%02d,%02d:%02d:%02d",
		      Time_PS->tm_year + 1900,
		      Time_PS->tm_mon + 1,
		      Time_PS->tm_mday,
		      Time_PS->tm_hour,
		      Time_PS->tm_min,
		      Time_PS->tm_sec );
	  }
	  String_PC = TmpStr_AC;
	  SendString_B = TRUE;
	  break;
      }

      case FN_STR2TIME: {
	  struct tm Time_S;
	  int yr,mo,da,hr,mn,se;

	  TmpInt_I = sscanf( arg->argv[0],
			    "%04d%*c%02d%*c%02d%*c%02d%*c%02d%*c%02d",
			    &yr, &mo, &da, &hr, &mn, &se );
	  if ( 6 > TmpInt_I )
	      TmpInt_I = 0;
	  else {
	      Time_S.tm_year = yr - 1900;
	      Time_S.tm_mon  = mo - 1;
	      Time_S.tm_mday = da;
	      Time_S.tm_hour = hr;
	      Time_S.tm_min  = mn;
	      Time_S.tm_sec  = se;
	      TmpInt_I = mktime( &Time_S );
	      if ( -1 == TmpInt_I )
		  TmpInt_I = 0;
	  }

	  SendInt_B = TRUE;

	  break;
      }

      case FN_STRFTIME: {
	time_t t;
	struct tm* ts;

	if ( arg->argc > 1 )
	    t = (int)arg->argv[1];
	else
	    t = time(NULL);
	ts = localtime( &t );
	TmpInt_I = strftime( TmpStr_AC, sizeof( TmpStr_AC ),
			     arg->argv[0], ts );
	SendString_B = TRUE;
	break;
      }

#if 0
      case FN_LOADFILE:
	CALL( FIL_LoadFile( arg->argv[0], &String_PC ) );
	if ( String_PC ) {
	    FPLSendString( String_PC );
	    free( String_PC );
	}
	else
	    FPLSendString( "" );
#endif


      case FN_ARGV: {
	  TmpInt_I = (int) (arg->argv[0]) - 1;

          if ((TmpInt_I <= Argc_I ) && (TmpInt_I>=0))
              ret = FPLSendString( Argv_PPC[TmpInt_I + 2] );
          else
              ret = FPLSendString( "" );
	  break;
      }


	/* Variables */

      case VR_ARGC:
	TmpInt_I = Argc_I;
	SendInt_B= TRUE;
	break;

      case VR_ERRNO:
	TmpInt_I = FplErr_I;
	SendInt_B= TRUE;
	break;

      case VR_VERSION:
	strcpy(TmpStr_AC,VERSION);
	String_PC = TmpStr_AC;
	SendString_B= TRUE;
	break;

      case VR_FPLVERSION: {
	long Version_I,Revision_I,rc;
	rc = fplSendTags( key, FPLSEND_GETVERSION, &Version_I,
                          FPLSEND_GETREVISION, &Revision_I, NULL );
        if ( rc )
	    RET_ERR( "Error getting FPL version", (int)rc );

	sprintf( TmpStr_AC, "%ld.%ld", Version_I, Revision_I );
	String_PC = TmpStr_AC;
	SendString_B= TRUE;
      }
	break;

      case VR_SAFEMODE:
	TmpInt_I = RunSafe_I ? 1 : 0;
	SendInt_B= TRUE;
	break; 
    }

    if ( SendInt_B )
	ret=FPLSendInt( TmpInt_I );

    if ( SendString_B )
	ret=FPLSendString( TmpStr_AC );

    return ret;
}
