/****************************************************************************
*
*  Local Defines
*
***************************************************************************/
#define INCL_DOSPROCESS
#define INCL_DOSFILEMGR
#define INCL_DOSMISC
#define INCL_DOSERRORS

/****************************************************************************
*
*  Local Includes
*
***************************************************************************/
#include <stdio.h>
#include <string.h>
#include <dirent.h>
#include <sys/stat.h>
#include <errno.h>

#include "stdcode.h"
#include "os.h"
#include "fplerr.h"

/****************************************************************************
*
*  Local Functions
*
***************************************************************************/

static t_RetCode os_CreateTime( struct stat* buf, t_FileTime* Time_PS );

/****************************************************************************
*
*  Function:    OS_FindFirst
*  Description: Find a filename
*
***************************************************************************/
t_RetCode OS_FindFirst( char* Pattern_PC, char* Flags_PC, char* DestBuf_PC )
{
    return OK;
}
    
/****************************************************************************
*
*  Function:    OS_FileInfo
*  Description: Obtain info struct about file
*
***************************************************************************/

t_RetCode OS_FileInfo( uchar* Filename_PC, t_FileInfo* Info_PS )
{
    int rc;
    struct stat buf;
 
    rc = stat( Filename_PC, &buf );

    if ( (-1 == rc) && ( ENOENT == errno ) ) {
        Info_PS->Missing_B = TRUE;
        return OK;
    }

    Info_PS->Missing_B = FALSE;
    Info_PS->Size_I = buf.st_size;
    CALL( os_CreateTime( &buf, &(Info_PS->Time_S) ) );

    return OK;
}

/****************************************************************************
*
*  Function:    os_CreateTime
*  Description: Create a t_FileTime struct from a FILESTATUS3 struct.
*
***************************************************************************/

static t_RetCode os_CreateTime( struct stat* buf, t_FileTime* Time_PS )
{
    struct tm* ts;

    ts = localtime( &(buf->st_mtime) );

    Time_PS->Year_I      = ts->tm_year + 1900;
    Time_PS->Month_I     = ts->tm_mon + 1;
    Time_PS->DayOfMonth_I= ts->tm_mday;
    Time_PS->Hour_I	 = ts->tm_hour;
    Time_PS->Minute_I    = ts->tm_min;
    Time_PS->Second_I    = ts->tm_sec;
    Time_PS->Time_I      = buf->st_mtime;

    return OK;
}

/****************************************************************************
*
*  Function:    OS_Milliseconds
*  Description: Report current millisecond count (real-time clock)
*
***************************************************************************/
t_RetCode	OS_Milliseconds( int* Ms_PI )
{
    struct timeval tp;
 
    gettimeofday( &tp, NULL );
 
    /* output new-row marker */
    *Ms_PI = tp.tv_usec;

    return OK;
}

/****************************************************************************
*
*  Function:    OS_Sleep
*  Description: Sleep a number of milliseconds
*
***************************************************************************/
t_RetCode OS_Sleep( int mSec_I )
{
#if 0
    DosSleep( mSec_I );
#endif
    return OK;
}


